/*******************************************************************************
 * File name: IntelJ3.cpp
 * Description: Intel Flash burner
 * Author: Shiva	
 * Revision history:
 * 2010/10/01	Create this file.
 * ****************************************************************************/

#include "IntelJ3.h"
using namespace std;

int sock;
int fast_mode = 0;
int debug_level = 0;  
FILE *log = stdout;                  /* log file */

/* send command to iceman */
int send_cmd(char cmd)
{
  char send_data[2],recv_data[2];
  send_data[0] = cmd;
#ifdef __MINGW32__
    send(sock, send_data, 1, 0);
#else
    write(sock, send_data, 1);
#endif
#ifdef __MINGW32__
    recv(sock, recv_data, 2, 0);
#else
    read(sock, recv_data, 2);
#endif
 //  fprintf (log, "%x %x\n",recv_data[0],recv_data[1]);
  if (recv_data[0] == cmd && recv_data[1] == 0) 
  {
    //fprintf (log, "Done.\n");
    return int(recv_data[1]);
  }
  else 
  {
    return -1;
  }
}

/* write word */
/* could use outw write word to target memory */
int outw(unsigned int address, unsigned int data)
{
  char send_data[10],recv_data[2];
  send_data[0] = 0x1a;
  send_data[2] = (char)((address & 0x000000FF) >> 0);
  send_data[3] = (char)((address & 0x0000FF00) >> 8);
  send_data[4] = (char)((address & 0x00FF0000) >> 16);
  send_data[5] = (char)((address & 0xFF000000) >> 24);
  send_data[6] = (char)((data & 0x000000FF) >> 0);
  send_data[7] = (char)((data & 0x0000FF00) >> 8);
  send_data[8] = (char)((data & 0x00FF0000) >> 16);
  send_data[9] = (char)((data & 0xFF000000) >> 24);
#ifdef __MINGW32__
    send(sock, send_data, 10, 0);
#else
    write(sock, send_data, 10);
#endif
#ifdef __MINGW32__
    recv(sock, recv_data, 2, 0);
#else
    read(sock, recv_data, 2);
#endif
  if (recv_data[0] == 0x1a) return int(recv_data[1]);
  else 
  {
    return -1;
  }
}

/* fast read */
/* could use fastin read mult-word from target memory */
int fastin(unsigned int address, unsigned int size, char * buffer)
{
  char send_data[10],recv_data[2048];
  unsigned int i;
  send_data[0] = 0x1c;
  send_data[2] = (char)((address & 0x000000FF) >> 0);
  send_data[3] = (char)((address & 0x0000FF00) >> 8);
  send_data[4] = (char)((address & 0x00FF0000) >> 16);
  send_data[5] = (char)((address & 0xFF000000) >> 24);
  send_data[6] = (char)((size & 0x000000FF) >> 0);
  send_data[7] = (char)((size & 0x0000FF00) >> 8);
  send_data[8] = (char)((size & 0x00FF0000) >> 16);
  send_data[9] = (char)((size & 0xFF000000) >> 24);
#ifdef __MINGW32__
    send(sock, send_data, 10, 0);
#else
    write(sock, send_data, 10);
#endif
#ifdef __MINGW32__
    recv(sock, recv_data, size + 2, 0);
#else
    read(sock, recv_data, size + 2);
#endif
  for (i = 0; i < size; i++) buffer[i] = recv_data[2+i];

  if (recv_data[0] == 0x1c) return 0;
  else 
  {
    return 0xffffffff;
  }
}
/* fast write */
/* could use fastout write mult-word from target memory */
int fastout(unsigned int address, unsigned int size, char * buffer)
{
  char send_data[2048],recv_data[2];
  unsigned int i;
  send_data[0] = 0x20;
  send_data[2] = (char)((address & 0x000000FF) >> 0);
  send_data[3] = (char)((address & 0x0000FF00) >> 8);
  send_data[4] = (char)((address & 0x00FF0000) >> 16);
  send_data[5] = (char)((address & 0xFF000000) >> 24);
  send_data[6] = (char)((size & 0x000000FF) >> 0);
  send_data[7] = (char)((size & 0x0000FF00) >> 8);
  send_data[8] = (char)((size & 0x00FF0000) >> 16);
  send_data[9] = (char)((size & 0xFF000000) >> 24);

  for (i = 0; i < size; i++) send_data[i+12] = buffer[i];
#ifdef __MINGW32__
    send(sock, send_data, size+12, 0);
#else
    write(sock, send_data, size+12);
#endif
#ifdef __MINGW32__
    recv(sock, recv_data, 2, 0);
#else
    read(sock, recv_data, 2);
#endif

  if (recv_data[0] == 0x20) return 0;
  else 
  {
    return 0xffffffff;
  }
}

/* read word */
/* could use inw read word from target memory */
unsigned int inw(unsigned int address)
{
  unsigned int data;
  char send_data[6],recv_data[6];
  send_data[0] = 0x1b;
  send_data[2] = (char)((address & 0x000000FF) >> 0);
  send_data[3] = (char)((address & 0x0000FF00) >> 8);
  send_data[4] = (char)((address & 0x00FF0000) >> 16);
  send_data[5] = (char)((address & 0xFF000000) >> 24);
#ifdef __MINGW32__
    send(sock, send_data, 6, 0);
#else
    write(sock, send_data, 6);
#endif
#ifdef __MINGW32__
    recv(sock, recv_data, 6, 0);
#else
    read(sock, recv_data, 6);
#endif
  data = ((((unsigned int)recv_data[2]<< 24) & 0xFF000000) |
          (((unsigned int)recv_data[3]<< 16) & 0x00FF0000) |
          (((unsigned int)recv_data[4]<< 8) & 0x0000FF00) |
          (((unsigned int)recv_data[5]<< 0) & 0x000000FF));

  if (recv_data[0] == 0x1b) return data;
  else 
  {
    return 0xffffffff;
  }
}

void check(unsigned data)
{
  if (data & 2)
  {
    fprintf(log, "\nFlash block locked.\n");
    fflush(log);
    terminate();
    exit(1);
  }
}

/* read flash id */
void Flash_ReadID(unsigned flash_type, unsigned base, unsigned BusWidth, unsigned* DDI1, unsigned* DDI2)
{
    if (flash_type == Intel_Type)
    {
        switch(BusWidth)
        {
        case FLASH_SINGLE: // 8 bit
            outw(base,0x90);
            *DDI1 = inw(base+0x00);
            *DDI2 = inw(base+0x01);
            outw(base,0xff);
            break;
        case FLASH_DOUBLE: // double 16 bit
            outw(base,0x9090);
            *DDI1 = inw(base+0x00);
            *DDI2 = inw(base+0x02);
            outw(base,0xffff);
            break;
        case FLASH_FOUR:// four 32 bit
            outw(base,0x00500050);
            outw(base,0x00900090);
            *DDI1 = inw(base+0x00);
            *DDI2 = inw(base+0x04);
            outw(base,0x00ff00ff);
            break;
        }
    }
}

void Flash_Reset(unsigned base, unsigned BusWidth)
{
    switch(BusWidth)
    {
    case FLASH_SINGLE:
        outw(base,0xf0);
        break;
    case FLASH_DOUBLE:
        outw(base,0xf0f0);
        break;
    case FLASH_FOUR:
        //outw(base,0xf0f0f0f0);
        outw(base,0xffffffff);
        break;
    }
}
void Flash_UnLock(unsigned flash_type, unsigned base, unsigned BusWidth)
{
    unsigned data;

    if (flash_type == Intel_Type)
    {
        // Only erase the first sector
        switch(BusWidth)
        {
        case FLASH_SINGLE: // 8 bit
            outw(base,0x50);
            outw(base,0x60);
            outw(base,0xd0);

            // read status register
            outw(base,0x70);
            data = inw(base);

            while (data != 0x80) {
               outw(base,0x70);
               data=inw(base);
            }

            break;
        case FLASH_DOUBLE: // double 16 bit
            outw(base,0x5050);
            outw(base,0x6060);
            outw(base,0xd0d0);

            // read status register
            outw(base,0x7070);
            data = inw(base);

            while (data != 0x8080) {
               outw(base,0x7070);
               data=inw(base);
            }

            break;

        case FLASH_FOUR:// four 32 bit
            outw(base,0x00500050);
            outw(base,0x00600060);
            outw(base,0x00d000d0);

            // read status register
            outw(base,0x00700070);
            data = inw(base);
            while (data != 0x00800080) {
               outw(base,0x00700070);
               data=inw(base);
            }
            break;
        }
    }
}


void Flash_Lock(unsigned flash_type, unsigned base, unsigned BusWidth)
{
    unsigned data;

    if (flash_type == Intel_Type)
    {
        // Only erase the first sector
        switch(BusWidth)
        {
        case FLASH_SINGLE: // 8 bit
            outw(base,0x60);
            outw(base,0x01);

            // read status register
            outw(base,0x70);
            data = inw(base);

            while (data != 0x80) {
               outw(base,0x70);
               data=inw(base);
            }

            break;
        case FLASH_DOUBLE: // double 16 bit
            outw(base,0x6060);
            outw(base,0x0101);

            // read status register
            outw(base,0x7070);
            data = inw(base);

            while (data != 0x8080) {
               outw(base,0x7070);
               data=inw(base);
            }

            break;

        case FLASH_FOUR:// four 32 bit
            outw(base,0x00600060);
            outw(base,0x00010001);

            // read status register
            outw(base,0x00700070);
            data = inw(base);
            while (data != 0x00800080) {
            //while ( (data&0xff) != 0x80) {
               outw(base,0x00700070);
               data=inw(base);
            }
            break;
        }
    }
}


void Flash_ChipErase(unsigned flash_type, unsigned base, unsigned BusWidth)
{
    unsigned data;
    unsigned time = 0;

    if (flash_type == Intel_Type)
    {
        // Only erase the first sector
        switch(BusWidth)
        {
        case FLASH_SINGLE: // 8 bit
            outw(base,0x20);
            outw(base,0xd0);

            outw(base,0xff);
            // read status register
            outw(base,0x70);
            data = inw(base);

            while (data != 0x80) {
               outw(base,0x70);
               data=inw(base);
            }
            outw(base,0xff);

            break;
        case FLASH_DOUBLE: // double 16 bit
            outw(base,0x2020);
            outw(base,0xd0d0);

            outw(base,0xffff);
            // read status register
            outw(base,0x7070);
            data = inw(base);

            while (data != 0x8080) {
               outw(base,0x7070);
               data=inw(base);
            }
            outw(base,0xffff);

            break;

        case FLASH_FOUR:// four 32 bit
            outw(base,0x00200020);
            outw(base,0x00d000d0);

            outw(base,0x00ff00ff);
            // read status register
            outw(base,0x00700070);
            data = inw(base);
            while (data != 0x00800080) {
            //while ( (data&0xff) != 0x80) {
               outw(base,0x00700070);
               data=inw(base);
               if (time > 1000) check(data);
               time++;
            }
            outw(base,0x00ff00ff);
            break;
        }
    }
}

// this function only use in FLASH_SINGLE
void Flash_ProgramByte(unsigned flash_type, unsigned base, unsigned address, UINT8 data)
{
    volatile unsigned int s_reg = 0;

    if (flash_type == Intel_Type)
    {
       outw(base,0x40);
       outw(address,data);

       outw(base,0x70);
       s_reg=inw(base);

        while ((s_reg&0xff) != 0x80) {
          outw(base,0x70);
          s_reg=inw(base);
       }
       outw(base,0xffffffff);
    }
}

// this function only use in FLASH_DOUBLE
void Flash_ProgramHalfWord(unsigned flash_type, unsigned base, unsigned address, UINT16 data)
{
    volatile unsigned int s_reg = 0;

    if (flash_type == Intel_Type)
    {
       outw(base,0x4040);
       outw(address,data);

       outw(base,0x7070);
       s_reg=inw(base);

       while ((s_reg&0xff) != 0x80) {
          outw(base,0x7070);
          s_reg=inw(base);
       }

       outw(base,0xffff);
    }
}
// this function only use in FLASH_FOUR
void Flash_ProgramWord(unsigned flash_type, unsigned base, volatile unsigned address, volatile unsigned data)
{
    volatile unsigned int s_reg = 0;
#ifndef FPGA 
    if (flash_type == Intel_Type)
#endif // #ifndef FPGA
    {
       outw(base, 0x00400040);
       outw(address, data);

       outw(base, 0x00700070);
       s_reg = inw(base);

       while (s_reg != 0x00800080)
       {
          outw(base, 0x00700070);
          s_reg = inw(base);
       }
       outw(base, 0x00ff00ff);
    }
}
/* only for intel flash */
/* max is 16 words in one time */
void Flash_ProgramMultiWord(unsigned flash_type, unsigned base, volatile unsigned address, unsigned char *buffer, unsigned size)
{
    volatile unsigned int s_reg = 0;
    volatile unsigned int block_addr = (address/BLK_SIZE)*BLK_SIZE;
    volatile unsigned int bytes = size % 4;
    volatile unsigned int word;
    volatile unsigned int time = 0;
 
    /* not word align */
    if (size % 4 != 0) size = size - bytes;
    word = (size/4) - 1;
    word = (word << 16) + word;

    if (flash_type == Intel_Type)
    {
       outw(block_addr, 0x00e800e8);
       outw(block_addr, word);
#if 0
       for(unsigned int i = 0; i < (size/4); i++) 
       {
         data = ((((unsigned int)buffer[4*i+3]<< 24) & 0xFF000000) |
                 (((unsigned int)buffer[4*i+2]<< 16) & 0x00FF0000) |
                 (((unsigned int)buffer[4*i+1]<< 8) & 0x0000FF00) |
                 (((unsigned int)buffer[4*i]<< 0) & 0x000000FF));
         outw(address + 4*i, data);
       }
#else
       fastout(address, size, (char*)buffer);
#endif
       outw(block_addr, 0x00d000d0);
       s_reg = 0;

       if (fast_mode == 0)
       {
         while (s_reg != 0x00800080)
         {
            outw(base, 0x00700070);
            s_reg = inw(base);
            if (time > 1000) check(s_reg);
            time++;
         }
       }
       outw(base, 0x00ff00ff);

       if (bytes)
       {
         word = 0;
         for (unsigned int j = 0; j <= bytes; j++) word = word | (buffer[size+j] << (j*8));
         Flash_ProgramWord (flash_type, base, address+size, word);
       }
    }
    else printf("not intel flash !!\n");
}

/* record the memory content to buffer */
int Record_content(unsigned int address, unsigned char* buffer, unsigned int size)
{
  char content[2048];
  unsigned int buffer_size, i, j = 0;

//  printf("record_content from 0x%x to 0x%x\n", address, address + size);
  while (size > 0)
  {
    if (size > 1024) buffer_size = 1024;
    else buffer_size = size;

    fastin(address, buffer_size, content);
    for (i = 0; i < buffer_size; i++)
    {
      buffer[j++] = (unsigned char)content[i];
    }
    size = size - buffer_size;
    address = address + buffer_size;
  }
  return 0;
}

/* burn image to flash */
int Burn_image(unsigned flash_type, unsigned int address, unsigned char* start, unsigned int size, unsigned int base)
{
  unsigned int end = address + size;
  unsigned int data;
  unsigned int buf_size;
  unsigned char temp[64];

  while (address < end)
  {
    if (flash_type == Intel_Type)
    {
      buf_size = ((end - address) > 64) ? 64 : (end - address);
      for (unsigned i = 0; i < buf_size; i++) temp[i] = *(start++);
      Flash_ProgramMultiWord(flash_type, base, address, temp, buf_size);
      address = address + buf_size;
    }
    else
    { 
      for (unsigned i = 0; i < 4; i++) temp[i] = *(start++);

      data = ((((unsigned int)temp[3]<< 24) & 0xFF000000) |
              (((unsigned int)temp[2]<< 16) & 0x00FF0000) |
              (((unsigned int)temp[1]<< 8) & 0x0000FF00) |
              (((unsigned int)temp[0]<< 0) & 0x000000FF));

      Flash_ProgramWord(flash_type, base, address, data);
      address = address + 4;
    }

    if (address % 8192 == 0)
    {
      fprintf(log, ".");
      fflush(log);
    }
  }
  return 0;
}
/* verify after burn */
/* return 1 if varify fail */
int Verification(unsigned int address, unsigned char* start, unsigned int size)
{
  char content[2048];
  unsigned int buffer_size, i;

  while (size > 0)
  {
    if (size > 1024) buffer_size = 1024;
    else buffer_size = size;
    fastin(address, buffer_size, content);
    for (i = 0; i < buffer_size; i++)
    {
      //printf("%x: %x | %x\n", address + i, (unsigned char)content[i], *(start));
      if ((unsigned char)content[i] != *(start++)) return 1;
    }
    size = size - buffer_size;
    address = address + buffer_size;
  }
  return 0;
}

void xc5(void)
{
  if (inw(0x90100018) != 0x10080000) outw(0x90100018, 0x10080000);
  if (inw(0x90200000) != 0x10000052) outw(0x90200000, 0x10000052);
  if (inw(0x90100010) != 0x00080000) outw(0x90100010, 0x00080000);

  outw(0x90200004, 0x00151151);
  outw(0x90300000, 0x00011312);
  outw(0x90300004, 0x00480180);
  outw(0x90300008, 0x00002326);
}
/* for 24 bit platform */
void p24(void)
{
  if (inw(0xe00018) != 0x10080000) outw(0xe00018, 0x10080000);
  if (inw(0xe01000) != 0x10000052) outw(0xe01000, 0x10000052);
  if (inw(0xe00010) != 0x00080000) outw(0xe00010, 0x00080000);

  outw(0xe01004, 0x00151151);
}

void ag301p(void)
{
  if (inw(0x90c00018) != 0x10080000) outw(0x90c00018, 0x10080000);
  if (inw(0x90200000) != 0x10000052) outw(0x90200000, 0x10000052);
  if (inw(0x90c00010) != 0x00080000) outw(0x90c00010, 0x00080000);
}

void ag101(void)
{
  unsigned int data;

  outw(0x90200004, 0x00151151);
  outw(0x90100018, 0x100b0000);
  data = inw(0x98100030);
  data = data | 0x00010000;
  outw(0x98100030, data);
  data = inw(0x98100030);
  data = data | 0x00e00000;
  outw(0x98100030, data);
  data = inw(0x98100028);
  data = data | 0x00000010;
  outw(0x98100028, data);
  data = inw(0x9810004c);
  data = data | 0x0007f000;
  outw(0x9810004c, data);
  outw(0x90300000, 0x00011312);
  outw(0x90300004, 0x00480180);
  outw(0x90300008, 0x00002326);
  outw(0x9030000c, 0x00000010);

  outw(0x90300010, 0x00001100);
  outw(0x90300014, 0x00000000);
  outw(0x90300018, 0x00000000);
  outw(0x9030001c, 0x00000000);
  outw(0x90300010, 0x00001000);

  outw(0x90200008, 0x10000052);
  outw(0x9020000c, 0x000ff3ff);
}

unsigned int usage(void)
{
  printf( "--port:\t\tSpecific port number to connect with ICEman (default to 2354)\n");
  printf( "--base:\t\tSpecific flash base address (default to 0x8040000)\n");
  printf( "--addr:\t\tSpecific flash target address to write (default to 0x8040000)\n");
  printf( "--image:\tSpecific image name to burn\n");
  printf( "--preserve:\tSpecific to preserve the content in first and last erase blocks\n");
  printf( "--log:\t\tSpecific the log file to store output message (default to stdout)\n");
  printf( "--reset-target:\ttarget reset.\n");
  printf( "--target:\tSpecific board type to map ROM start address to 0x0 (xc5/ag101/ag101p_16mb)\n");
  printf( "--verify:\tVerify after flash burning\n");
  printf( "--version:\tShow flash burning version\n");
  printf( "--unlock:\tUnlock flash before burning\n");
  printf( "--lock:\t\tLock flash after burning\n");
  printf( "--fast:\t\tBurning flash in fast mode\n");
  printf( "--erase-all:\tErase entire flash before burning\n");
  printf( "--help:\t\tThe usage for IntelJ3 Burner\n");
  return 0;
}

/* Need send terminate to ICEman before quit Burner */
void terminate(void)
{
  char send_data[2];
  send_data[0] = 4;
#ifdef __MINGW32__
  send(sock, send_data, 1, 0);
#else
  write(sock, send_data, 1);
#endif
}

void handle_int(int signo)
{
  terminate();
  fprintf(log, "\n!! burn not done yet !!\n");
  fflush(log);
  exit(0);
}

int main(int argc, char **argv)
{
  int c = 0;
  unsigned int port = 2354;         /* default connecting port */
  unsigned int base = 0x80400000;    /* flash base address */
  unsigned int address = 0x80400000; /* default write to flash base */
  unsigned int size;                 /* image size */
  unsigned int erase_size;           /* erase size */
  unsigned int end;                  /* binary end address */
  unsigned int pt, i, j;
  unsigned int id1, id2;             /* flash id */ 
  unsigned int first_size, last_size;             
  unsigned char* flash = NULL;       /* image pointer */
  unsigned char* start = NULL;       /* image pointer */
  unsigned char first[BLK_SIZE], last[BLK_SIZE];
  FILE *image = NULL;                /* image to burn */
  char flash_type = Intel_Type;
  char cmd = 0;
  char ip[20] = "127.0.0.1";
  char *name = new char[80];
  int preserve = 0;
  int remap = 0;
  int specify_base = 0;
  int specify_address = 0;
  int xc5_enable = 0;
  int ag301p_enable = 0;
  int ag101_enable = 0;
  int p24_enable = 0;
  int verify = 0;
  int lock = 0;
  int unlock = 0;
  int erase_all = 0;
  struct sockaddr_in server_addr;  

  /* parameter handler */
  while(1)
  {
    int option_index;
    c = getopt_long(argc, argv, "uzgyvVtoehl:x:k:p:a:b:f:i", long_option, &option_index);
    if (c == EOF) break;

    switch (c)
    {
      case 'u':
	erase_all = 1;
	break;
      case 'z':
	lock = 1;
	break;
      case 'g':
	fast_mode = 1;
	break;
      case 'y':
	unlock = 1;
	break;
      case 'v':
	verify = 1;
	break;
      case 'l':
	if ((log = fopen(optarg, "w"))== NULL) {
          fprintf (log, "Error: can't open file %s\n", optarg);
          fflush(log);
          exit(1);
        }
        break;
      case 's':
        fprintf (log, "Branch%s\n", BRANCH_NAME);
        fprintf (log, "%s\n", COMMIT_ID);
        fflush(log);
        exit(0);
      case 'V':
        fprintf (log, "IntelJ3 Burner BUILD_ID: %d\n", BUILD_ID);
        fflush(log);
        exit(0);
	break;
      case 'x':
        if (strncmp(optarg, "xc5", 6) == 0)
        {
          xc5_enable = 1;
          remap = 1;
        } else if (strncmp(optarg, "ag101", 6) == 0)
        {
  	  ag101_enable = 1;
          remap = 1;
        } else if (strncmp(optarg, "ag301p", 6) == 0)
        {
  	  ag301p_enable = 1;
          remap = 1;
        } else if (strncmp(optarg, "ag101p_16mb", 6) == 0)
        {
  	  p24_enable = 1;
          remap = 1;
        }
        break;
      case 't':
        cmd = RESET_TARGET;
        break;
      case 'o':
        cmd = RESET_HOLD;
        break;
      case 'e':
        preserve = 1;
        break;
      case 'k':
        sscanf(optarg,"%s", ip);
        break;
      case 'p':
        port = strtol(optarg, NULL, 0);
        break;
      case 'a':
        sscanf(optarg,"0x%x", &address);
        specify_address = 1;
        break;
      case 'b':
        sscanf(optarg,"0x%x", &base);
        specify_base = 1;
        break;
      case 'f':
        if (strncmp(optarg, "mx", 6) == 0)
        {
          flash_type = MX_Type;
        } else if (strncmp(optarg, "sst", 6) == 0)
        {
          flash_type = SST_Type;
        } else if (strncmp(optarg, "intel", 6) == 0)
        {
          flash_type = Intel_Type;
        } else if (strncmp(optarg, "amd", 6) == 0)
        {
          flash_type = AMD_Type;
        }
        break;
      case 'i':
        if (strncmp(optarg, "/cygdrive/", 10) == 0)
        {
          name[0] = optarg[10];
          name[1] = ':';
          name[2] = '\0';
          strcat(name, optarg + 11);
        }
        else name = optarg;
	if ((image = fopen(name, "rb"))== NULL) {
                fprintf (log, "Error: can't open file %s\n", optarg);
                fflush(log);
                exit(1);
        }
        break;
      case 'h':
      case '?':
      default:
        usage();
        exit(0);
        break;
    }
  }

  fprintf (log, "IntelJ3 Burner BUILD_ID: %d\n", BUILD_ID);
  fflush(log);

  if (remap && (specify_address == 0) ) address = 0;
  if (remap && (specify_base == 0) ) base = 0;

  /* initial socket */
#ifdef __MINGW32__
  WSADATA wsadata;
  if (WSAStartup(MAKEWORD(1,1), &wsadata) == SOCKET_ERROR) {
    if (debug_level >= 3)
    {
      fprintf(log, "Error creating socket.");
      fflush(log);
      fflush(stdout);
    }
  }
#endif

  if ((sock = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
     fprintf(log, "cant create socket\n");
     fflush(log);
     exit(1);
  }
  server_addr.sin_family = PF_INET;     
  server_addr.sin_port = htons(port);   
#ifdef __MINGW32__
  server_addr.sin_addr.s_addr = inet_addr( ip );
#else
  inet_aton(ip, &server_addr.sin_addr);
#endif 
  memset(&(server_addr.sin_zero),0,8); 
  if (connect(sock, (struct sockaddr *)&server_addr,
                    sizeof(struct sockaddr)) == -1) 
  {
    fprintf(log, "connecting fail !!\n");
    fflush(log);
    exit(1);
  }

#ifdef __MINGW32__
  Sleep (200);
#else
  usleep (200000);
#endif
  send_cmd(RESET_AICE);

#if 0
  send_cmd(RESET_AICE);
  terminate();
  exit(0);
#endif
  if (xc5_enable)
  {
    send_cmd(HOLD_CORE);
    xc5();
  }
  if (ag101_enable)
  {
    send_cmd(HOLD_CORE);
    ag101();
  }
  if (ag301p_enable)
  {
    send_cmd(HOLD_CORE);
    ag301p();
  }

  if (p24_enable)
  {
    send_cmd(HOLD_CORE);
    p24();
  }
  if (cmd != 0)
  {
    send_cmd(cmd);
    terminate();
    exit(0);
  }

  /* get burn image content */
  if (image == NULL)
  {
    fprintf (log, "Error: Burn image have to specify.\n");
    fflush(log);
    exit(1);
  }

  fseek(image, 0L, SEEK_END);
  size = (ftell(image));
  fseek(image, 0L, SEEK_SET);

  if ((flash = (UINT8*)malloc(size * sizeof(UINT8))) == NULL) {
     fprintf(log, "Error: can't allocate memory (%d bytes) for flash file\n", size);
     fflush(log);
     return 0;
  }
  fread(flash, size, 1, image);
  start = flash;

  /* handle signal */
  signal(SIGTERM, handle_int);
  signal(SIGINT, handle_int);
#ifndef __MINGW32__
  signal(SIGKILL, handle_int);
#endif
  /* calculate address */
  end = address + size;
  pt = (address/BLK_SIZE) * BLK_SIZE;
  erase_size = ((size/BLK_SIZE) + 1) * BLK_SIZE;

  /* check device ID */
  Flash_ReadID(flash_type, base, FLASH_FOUR, &id1, &id2);
  if (id1 == 0x890089 && id2 == 0x180018) {
     fprintf(log, "burn data to intel flash\n");
     fflush(log);
  }
  else
  {
    terminate();
    fprintf(log, "Not intel flash !!\n");
    fflush(log);
    return 0;
  }
  /* unlock flash */
  if (unlock)
  {
    fprintf(log, "unlocking");
    fflush(log);
    for (i = 0, j = pt; i < erase_size/BLK_SIZE; i++)
    {
      fprintf(log, ".");
      fflush(log);
      Flash_UnLock(flash_type, j, FLASH_FOUR);
      j += BLK_SIZE;
    }
    fprintf(log, "\n");
    fflush(log);
  }
  /* record the content in first and last erase block */
  if (preserve)
  {
    first_size = address - pt;
    last_size = erase_size - size - first_size;

    if (first_size)
    {
      Record_content(pt, first, first_size);
    }
    if (last_size)
    {
      Record_content(end, last, last_size);
    }
  }
  /* block erase */
  if (erase_all)
  {
    pt = base;
    erase_size = FLASH_SIZE;
    if (p24_enable) erase_size = 0x100000;
  }
  fprintf(log, "erase from address = %x\n", pt);
  fflush(log);
  for (i = 0, j = pt; i < erase_size/BLK_SIZE; i++)
  {
    fprintf(log, "erasing block %d (0x%x ~ 0x%x)\n", i+1, j, j + BLK_SIZE);
    fflush(log);
    Flash_ChipErase(flash_type, j, FLASH_FOUR);
    j += BLK_SIZE;
  }
  if (debug_level >= 3)
  {
    fprintf(log, "address = %x, base = %x, flash type = %d, size = %x\n", address, base, flash_type, size); 
    fflush(log);
  }
  fprintf(log, "burn flash from 0x%x to 0x%x\n", address ,end);
  fflush(log);
  fprintf(log, "burning");
  fflush(log);
  /* burn image */
  Burn_image(flash_type, address, flash, size, base);
  /* burn first and last block content back to flash */

  if (preserve)
  {
    if (first_size)
      Burn_image(flash_type, pt, first, first_size, base);
    if (last_size)
      Burn_image(flash_type, end, last, last_size, base);
  }
  /* verify */
  if (verify)
  {
    fprintf(log, "\nVerifying...\n");
    fflush(log);
    if (Verification(address, start, size))
    {
      fprintf(log, "Verify fail !!!\n");
      fprintf(log, "Please reburn again.\n");
      fflush(log);
    }
    else 
    {
      fprintf(log, "\nVerify success.\n");
      fprintf(log, "Flash burning done.\n");
      fflush(log);
    }
  }
  else
  {
    fprintf(log, "\nFlash burning done.\n");
    fflush(log);
  } 

  /* lock flash */
  if (lock)
  {
    fprintf(log, "locking");
    fflush(log);
    for (i = 0, j = pt; i < erase_size/BLK_SIZE; i++)
    {
      fprintf(log, ".");
      fflush(log);
      Flash_Lock(flash_type, j, FLASH_FOUR);
      j += BLK_SIZE;
    }
    fprintf(log, "\n");
    fflush(log);
  }
 
  /* send terminate signal */
  terminate();
  return 0;
}

